#!/usr/bin/python

###########################################
#                                         #
#      |'\     |   |         /''\   /  |  #
#      |./    .|  .|         |  |   \__|  #
#  /'\ |  /\ / | / | /| |'   |  |      |  #
#  \_/ |  \/ \_| \_| \_ |    \../ o    |  #
# \_/     www.perli.net/projekte/gpodder  #
#                                         #
###########################################


# gPodder 0.4
# ===========
# Copyright (c) 2005 Thomas Perl <thp@perli.net>
# Released on 06.09.2005 under the GNU GPL
#
# Please read the README file or go to the 
# gPodder website for more information.



### IMPORTS ###

import os        # Operating System utils
import os.path   # Path functionality (homedir, ..)
import re        # Regular expressions
import string    # String manipulation utils
import md5       # Misused for random name generator
import threading # Will we finally be able to be threaded?
import popen2    # Process open functions for wget
import time      # For sleeping purposes ;)

import pygtk     # PyGTK bindings (GTK+ for Python)
import gtk       # Well.. GTK stuff
import gtk.glade # Glade stuff (UI layout)



## GLOBAL DEFINITIONS ##

version = "0.4"
gpodderpath = os.path.expanduser( "~/.gpodder/")
configpath = gpodderpath
configfile = configpath + "gpodder.conf"

# set sane defaults =)
downloadpath = gpodderpath + "downloads/"
playerappcmd = "xmms"

temppath = gpodderpath + "temp/"

wget_pid = 0
show_downloaded = False


## CONFIGURATION UTILITY FUNCTIONS ##

# TODO: make the configuration handling more coolish, mainly try to 
#       make it xml-based (or at least object oriented ;).. well then
#       here we go with the current implementation :)    -- thp

# loadConfig: loads the configuration from a filename
def loadConfig( filename):
	global downloadpath
	global playerappcmd
	# TODO: add error checking into this function so it won't fuck up =)
	
	try:
		lf = open( filename, "r")
	except:
		return 1
	# end try-except
	
	if lf.readline() == "gPodder-"+version+"-config" + "\n":
		downloadpath = string.strip( lf.readline())
		playerappcmd = string.strip( lf.readline())
		#print "seems like i loaded the config :) congratulate me!"
		#print "downloadpath is now: " + downloadpath
		#print "playerappcmd is now: " + playerappcmd
	else:
		#print "well well.. configuration loading failed :/"
		print "** CONFIGURATION LOADING FAILED **"
	# end if
	lf.close()

	return 0
# end loadConfig()

# saveConfig: saves the configuration to a filename
def saveConfig( filename):
	global downloadpath
	global playerappcmd
	sf = open( filename, "w")
	sf.write( "gPodder-"+version+"-config" + "\n")
	sf.write( downloadpath + "\n")
	sf.write( playerappcmd + "\n")
	sf.close()
# end saveConfig()



## PODCAST UTILITY FUNCTIONS ##

# showMessage: show message to user to inform about something
def showMessage( message):
	# event handler for ok button
	def click_okbutton( widget):
		messagebox.destroy()
	# end click_okbutton()
	
	gladefile_message = gtk.glade.XML( "gpodder.glade", "messageBox")
	messagebox = gladefile_message.get_widget( "messageBox")
	label = gladefile_message.get_widget( "lblMessage")
	label.set_text( message)

	# -- insert your favourite connections joke here --
	message_connections = {
		"on_btnOK_clicked": click_okbutton
	}

	gladefile_message.signal_autoconnect( message_connections)
# end showMessage()

# urlToHostDesc: extract hostname from full URL
def urlToHostDesc( url):
	url1 = url[url.find( "://")+3:]
	url2 = url1[:url1.find( "/")]
	url3 = url[:url.find( ":")]
	return url2 + " (using " + url3 + ")"
# end urlToHostDesc()

# podcastPlayer: plays the download directory in selected player app
def podcastPlayer( filename = None):
	global downloadpath
	
	if filename == None:
		filename = downloadpath
	# end if
	
	os.system( playerappcmd+" \"" + filename + "\" &")
# end podcastPlayer()

# podcastFinder: get podcast urls from xml input and return array
def podcastFinder( input):
	casts = []
	input = input.replace( '&quot;', '"')
	# create regular expression
	p = re.compile('.*url=["\']([^"\']*)["\'].*')
	iter = p.finditer( input)
	
	# search for all urls
	for m in iter:
		basename = os.path.basename( m.group(1))
		#print '(podcastfinder) found ' + basename
		#updui()
		casts.append( m.group(1))
	
	# return that basename (filename)
	return casts
# end podcastFinder()

def getDownloadFilename( url):
	global downloadpath
	filename = os.path.basename( url)
	
	# strip question mark (and everything behind it)
	indexOfQuestionMark = filename.rfind( "?")
	if( indexOfQuestionMark != -1):
		filename = filename[0:indexOfQuestionMark]
	# end if
	
	downloadfilename = downloadpath+filename
	
	return downloadfilename.replace( "%20", " ")
# end getDownloadFilename()

# downloadFile: simply download a file to current directory
def downloadFile( url):
	global show_downloaded
	filename = getDownloadFilename( url)
	
	if show_downloaded == True:
		podcastPlayer( filename)
		return -1
	# end if
	
	# download it, and never overwrite anything =)
	result = downloadProcedure( url, filename, False)
	
	return result
# end downloadFile()

# getWebData: get an rss feed and save it locally, return content
def getWebData( url, force_update):
	filename = configpath + md5.new( url).hexdigest() + ".feed"

	# download it.. possibly re-download when force_update is True
	downloadProcedure( url, filename, force_update)

	# read the downloaded content and return it
	stream = open( filename, "r")
	return stream.read()
# end getWebData()

# thread function
def downloadThread( url, filename, readyEvent = None):
	global wget_pid
	
	global dlinfo_speed
	global dlinfo_percentage
	global dlinfo_result

	url = url.replace( "%20", " ")
	
	# first, download to tempfilename and on success move to filename
	tempfilename = temppath + os.path.basename( filename)
	#print "(downloadthread) downloading to tempfile: " + tempfilename
	command = "wget \""+url+"\" -O \""+tempfilename+"\""
	
	wgetp = popen2.Popen3( command, True)
	wget_pid = wgetp.pid
	reader = wgetp.childerr
	while wgetp.poll() == -1:
		line = reader.readline( 80)
		
		# the next line is debugging output from wget process
		#print line
		
		msg = string.strip( line)
		if msg.find("%") != -1:
			percentage = msg[(msg.find("%") - 2)] + msg[(msg.find("%") - 1)]
			speed = "unknown"
			p = re.compile('...\... .B\/s')
			iter = p.finditer( msg)
			for speed_string in iter:
				speed = string.strip(speed_string.group(0))
			
			dlinfo_speed = speed
			dlinfo_percentage = percentage
		# end if
	# end for line in reader
	
	# we wait for it to be finished, then rename when successful
	if wgetp.wait() == 0:
		os.rename( tempfilename, filename)
	# end if
	
	# reset wget pid because it is closed now..
	dlinfo_result = wgetp.poll()
	wget_pid = 0
	
	if readyEvent != None:
		readyEvent.set()
	# end if
# end downloadThread()

# downloadProcedure: gerneric implementation of downloading with gui
def downloadProcedure( url, filename, force_update):
	global dlinfo_speed
	global dlinfo_percentage
	global dlinfo_result
	
	url = url.replace( "%20", " ")
	
	dlinfo_speed = "initializing download..."
	dlinfo_percentage = "0"
	dlinfo_result = -1
	
	# check if file does not exist and download if necessary
	if( os.path.exists( filename) == False or force_update == True):
		wait_dialog_display( url, filename, dlinfo_speed)
		
		# set up the thread and start it in the background
		finished = threading.Event()
		argumente = ( url, filename, finished)
		mythread = threading.Thread( target=downloadThread, args=argumente )
		mythread.start()

		# wait for thread to be finished
		while finished.isSet() == False:
			finished.wait( 0.1)
			wait_dialog_update( url, filename, dlinfo_speed, dlinfo_percentage)
			updui()
		# end while
		wait_dialog_destroy()
	# end if
	
	# the error handling comes here..
	if dlinfo_result > 0:
		if dlinfo_result == 9:
			showMessage( "Download has been cancelled.")
		else:
			showMessage( "wget exited with status: " + str( dlinfo_result))
			print "*** THERE HAS BEEN AN ERROR WHILE DOWNLOADING **"
		# end if
	# end if
	
	# make the download result available to caller
	return dlinfo_result
# end downloadProcedure()

# getPodcastListFilename: return filename where podcastlist is saved
def getPodcastListFilename():
	return configpath + "podcasts.txt"
# end getPodcastListFilename()

# getPodcastList: load podcast list from local config file (podcasts.txt)
def getPodcastList():
	list = []
	filename = getPodcastListFilename()
	
	# if not existent, simply return an empty list (i.e. we start a new list)
	if( os.path.exists( filename) == False):
		return []
	
	# if it exists, go and read it!
	listfile = open( filename, "r")
	item = listfile.readline()
	while( item.strip() != ''):
		#print "(getpodcastlist) adding " + item.strip()
		list.append( item.strip())
		item = listfile.readline()
	listfile.close()
	return list
# end getPodcastList()

# putPodcastList: save podcast list to local config file (podcasts.txt)
def putPodcastList( list):
	filename = getPodcastListFilename()
	#print "(putpodcastlist) writing to " + filename
	
	listfile = open( filename, "w")
	
	for item in list:
		#print "(putpodcastlist) storing " + item
		listfile.write( item + "\n")
	
	listfile.close()
# end putPodcastList()

# getPodcastListModel: create a list model from our podcast list
def getPodcastListModel( feeds_list, force_update):
	global show_downloaded
	ts = gtk.TreeStore(str)
	
	for f in feeds_list:
		if string.strip( f) == "":
			continue
		# end if
		
		# append feed to feedlist
		feedit = ts.append( None, [f])

		# find subitems ("episodes")
		episodes = getWebData( f, force_update)
		casts = podcastFinder( episodes)
		
		for c in casts:
			# append subitems ("episode")
			if os.path.exists( getDownloadFilename( c)) == show_downloaded:
				ts.append( feedit, [c])
			# end if
		# end for
	
	return ts
#end getPodcastListModel()



## GUI HELPER FUNCTIONS ##

# wait_dialog_display: show our neat download dialog
def wait_dialog_display( url = "initializing...", filename = "initializing...", speed = "n/a"):
	global waitdlg
	global waitdlg_labelURL
	global waitdlg_labelFilename
	global waitdlg_labelSpeed
	global waitdlg_waitProgress
	
	gladewait = gtk.glade.XML("gpodder.glade", "waitDialog")
	waitdlg = gladewait.get_widget( "waitDialog")

	# the cancel event handler
	def waiter_cancel( widget):
		global wget_pid
		
		if wget_pid != 0:
			os.system( "kill -9 " + str( wget_pid))
		# end if
	# end waiter_cancel()

	# imagine the waiter has got connections to the cook.. (?!?!?)
	waiter_connections = {
	   "on_btnCancel_clicked": waiter_cancel
	}
	gladewait.signal_autoconnect( waiter_connections)
	
	# get widgets for later use
	waitdlg_labelURL = gladewait.get_widget( "labelURL")
	waitdlg_labelFilename = gladewait.get_widget( "labelFilename")
	waitdlg_labelSpeed = gladewait.get_widget( "labelSpeed")
	waitdlg_waitProgress = gladewait.get_widget( "waitProgress")
	
	# set "first-use" values
	waitdlg_labelURL.set_markup( "<b>" + urlToHostDesc( url) + "</b>")
	waitdlg_labelFilename.set_markup( "<b>" + os.path.basename( filename) + "</b>")
	waitdlg_labelSpeed.set_markup( "<b>" + speed + "</b>")
	
	updui()
#end wait_dialog_display()

# wait_dialog_update: update already visible "please wait" dialog with new text
def wait_dialog_update( url = "(unknown url)", filename = "(unknown filename)", speed = "n/a", progress = "0"):
	global waitdlg
	global waitdlg_labelURL
	global waitdlg_labelFilename
	global waitdlg_labelSpeed
	global waitdlg_waitProgress
	
	# set new values for our neat widgets
	waitdlg_labelURL.set_markup( "<b>" + urlToHostDesc( url) + "</b>")
	waitdlg_labelFilename.set_markup( "<b>" + os.path.basename( filename) + "</b>")
	waitdlg_labelSpeed.set_markup( "<b>" + speed + "</b>")
	
	# make it happen - the progressbar-o-neat stuff
	waitdlg_waitProgress.set_text( progress + "%")
	try:
		waitdlg_waitProgress.set_fraction( float( progress)/100.00)
	except:
		# this happens when crazy ppl insert spaces (%20) into urls
		waitdlg_waitProgress.set_fraction( 0)
		waitdlg_waitProgress.set_text( "0%")
	# end try-except
	#waitdlg.set_position(gtk.WIN_POS_CENTER_ALWAYS)
	updui()
# end wait_dialog_update()

# wait_dialog_destroy: destroy "please wait" dialog (close and hide)
def wait_dialog_destroy():
	global waitdlg
	waitdlg.destroy()
# end wait_dialog_destroy()

# updui: update gui.. this should be removed as soon as we do threaded operations
def updui():
	while( gtk.events_pending()):
		gtk.main_iteration( False)
# end updui()



## GUI EVENT HANDLERS ##

# gui_startup: all that happens on gui startup, init stuff and so on
def gui_startup():
	global feeds
	
	# load local podcast list
	feeds = getPodcastList()
	
	# setup our treeview widget
	column = gtk.TreeViewColumn( "Available Feeds")
	tree.append_column( column)
	cell = gtk.CellRendererText()
	column.pack_start( cell, True)
	column.add_attribute( cell, 'text', 0)
	
	# fill our treeview with data, not forcing re-download
	gui_redo_tree( False)
# end gui_startup()

# gui_redo_tree: re-fill treeview with data from gui
def gui_redo_tree( force_update = False, show_downloaded = False):
	global tree
	global feeds
	global model
	
	# will redo the tree of available feeds
	model = getPodcastListModel( feeds, force_update)
	tree.set_model( model)
	tree.expand_all()
	tree.queue_draw()
# end gui_redo_tree()

# gui_exit: clean up before exiting. (including exiting itself)
def gui_exit( widget, event):
	global feeds

	# make sure wget exists if it is still active
	if wget_pid != 0:
		os.system( "kill -9 " + str( wget_pid))
	# end if
	
	# saving settings and stuff
	#print "(guiexit) saving podcast list"
	putPodcastList( feeds)
	
	# exit now - print for debugging purposes
	#print "(guiexit) yay.. exiting now"
	gtk.main_quit()
# end gui_exit()

# gui_tree_row: download selected item
def gui_tree_row( widget, event, data):
	gui_btn_download( widget)
# end gui_row()

# gui_btn_add: handler for add button (thanks to joseph)
def gui_btn_add( widget):
	def feed_add(widget):
		feedstr = feedURL.get_text()
		feeds.append( feedstr)
		addfeed.destroy()
		gui_redo_tree( False)
	#end of feed_add
	
	def feed_cancel(widget):
		addfeed.destroy()
	#end of feed_cancel
	
	# create window, display       
	gladeaddfeed = gtk.glade.XML( "gpodder.glade", "addDialog")
	addfeed = gladeaddfeed.get_widget( "addDialog")
	feedURL = gladeaddfeed.get_widget( "entryURL")
	
	# connections, ya i got connections
	addconnections = {
	   "on_btnAdd_clicked": feed_add,
	   "on_btnCancel_clicked": feed_cancel
	}
	
	gladeaddfeed.signal_autoconnect( addconnections)
# end gui_btn_add()

# gui_btn_edit: edits a podcast in the list.. should be made better
def gui_btn_edit( widget):
	def feed_editor( url):
		def feed_edit(widget):
			feedstr = feedURL.get_text()
			feeds.remove( url)
			feeds.append( feedstr)
			addfeed.destroy()
			gui_redo_tree( True)
		#end of feed_edit
		
		def feed_cancel(widget):
			addfeed.destroy()
		#end of feed_cancel
		
		if url not in feeds:
			showMessage( "you can only edit feeds, not episodes!")
			return 0
		# end if
		
		# create window, display       
		gladeaddfeed = gtk.glade.XML( "gpodder.glade", "addDialog")
		addfeed = gladeaddfeed.get_widget( "addDialog")
		feedURL = gladeaddfeed.get_widget( "entryURL")
		
		feedURL.set_text( url)
		
		# connections, ya i got connections
		addconnections = {
		   "on_btnAdd_clicked": feed_edit,
		   "on_btnCancel_clicked": feed_cancel
		}
		
		gladeaddfeed.signal_autoconnect( addconnections)
	# end feed_editor()

	gui_iter_selected( feed_editor)
# end gui_btn_edit()

# gui_btn_remove: removing a feed from the list and updating list itself
def gui_btn_remove( widget):
	def remove_feeds( url):
		feeds.remove( url)
	# end remove_Feed()
	
	# iterate over all selected feeds
	gui_iter_selected( remove_feeds)
	
	gui_redo_tree()
# end gui_btn_remove()

# gui_btn_update: gets invoked when update button is clicked (force update)
def gui_btn_update( widget):
	#print "(guibtnupdate) updating podcast feed list"
	gui_redo_tree( True)
# end gui_btn_update()

# gui_btn_setup: shows setup screen and lets user configure things =)
def gui_btn_setup( widget):
	def action_ok( widget):
		global downloadpath
		global playerappcmd
		
		downloadpath = edit_download.get_text()
		playerappcmd = edit_playerapp.get_text()
		#print "downloadpath = " + downloadpath
		#print "playerappcmd = " + playerappcmd
		
		# we gonna create that dir so we can download into it
		try:
			os.mkdir( downloadpath)
		except:
			print "(info) using download dir " + downloadpath
		# end try-except
		
		setup_dialog.destroy()
	# end action_ok()
	
	def action_cancel( widget):
		#print "CANCEL clicked - will not save anything.."
		setup_dialog.destroy()
	# end action_cancel()
	
	gladefile_setup = gtk.glade.XML( "gpodder.glade", "setupDialog")
	setup_dialog = gladefile_setup.get_widget( "setupDialog")
	
	# the "download directory" entry field
	edit_download = gladefile_setup.get_widget( "editDownload")
	edit_download.set_text( downloadpath)

	# the "player app" entry field
	edit_playerapp = gladefile_setup.get_widget( "editPlayerApp")
	edit_playerapp.set_text( playerappcmd)
	
	# we "set up" connections.. hehe
	setupconnections = {
	   "on_btnOK_clicked": action_ok,
	   "on_btnCancel_clicked": action_cancel
	}
	
	gladefile_setup.signal_autoconnect( setupconnections)
# end gui_btn_setup()

# gui_iter_selected: iterate over all selected rows in treeview
def gui_iter_selected( callback_function):
	sel = tree.get_selection().get_selected_rows()
	model = sel[0]
	paths = sel[1]
	
	for path in paths:
		it = model.get_iter( path)
		try:
			url = model.get_value( it, 0)
			callback_function( url)
		except:
			continue
		# end try-except
	# end for
# end gui_iter_selected()

# gui_btndownload_rename: rename btnDownload, depending on mode
def gui_btndownload_rename():
	global show_downloaded
	global img_btn_download
	global lbl_btn_download
	global btn_update

	btn_update.set_child_visible( not show_downloaded)
	
	if show_downloaded:
		lbl_btn_download.set_label( "Open in media player")
		img_btn_download.set_from_stock( gtk.STOCK_MEDIA_PLAY, gtk.ICON_SIZE_BUTTON)
	else:
		lbl_btn_download.set_label( "Download selection")
		img_btn_download.set_from_stock( gtk.STOCK_GOTO_BOTTOM, gtk.ICON_SIZE_BUTTON)
	# end if
# end gui_btndownload_rename()

# gui_btn_show_available: show available downloads (not yet downloaded)
def gui_btn_show_available( widget):
	btn_showDownloaded.set_active( not btn_showAvailable.get_active())
	global show_downloaded
	show_downloaded = btn_showDownloaded.get_active()
	gui_btndownload_rename()
	gui_redo_tree( False)
# end gui_btn_show_available()

# gui_btn_show_downloaded: show downloaded podcasts (already downloaded)
def gui_btn_show_downloaded( widget):
	btn_showAvailable.set_active( not btn_showDownloaded.get_active())
	global show_downloaded
	show_downloaded = btn_showDownloaded.get_active()
	gui_btndownload_rename()
	gui_redo_tree( False)
# end gui_btn_show_downloaded()

# gui_btn_download: gets invoked when download button is clicked (well.. download!)
def gui_btn_download( widget):
	def handle_download( url):
		if url in feeds:
			showMessage( "Please click on an episode, not a feed..")
		else:
			downloadFile( url)
		# end if
	# end handle_download()
	
	# iterate over all selected fields
	gui_iter_selected( handle_download)
	
	# after successful download, update our tree (BUT ONLY IF NOT SHOWING DOWNLOADED PODCASTS ;)
	global show_downloaded
	if show_downloaded == False:
		gui_redo_tree( False)
	# end if
# end gui_btn_download()

# gui_btn_play: open xmms and play already downloaded podcasts
def gui_btn_play( widget):
	#print "(guibtnplay) playing in xmms"
	mythread = threading.Thread( target=podcastPlayer )
	mythread.start()
# end gui_btn_play()



## THE JUICE (AKA PUBLIC VOID MAIN) ##

# load config to init our variables right =)
loadConfig( configfile)

try:
	os.mkdir( configpath)
except:
	print "(info) using config dir " + configpath

try:
	os.mkdir( downloadpath)
except:
	print "(info) using download dir " + downloadpath

try:
	os.mkdir( temppath)
except:
	print "(info) using temporary dir " + temppath

# create the window from gladefile
glade = gtk.glade.XML("gpodder.glade", "gpodderWindow")
win = glade.get_widget( "gpodderWindow")
win.set_title( win.get_title().replace( "VV", version))

# connect gladefile with event handlers
connections = {
   "on_gpodderWindow_delete_event": gui_exit,
   "on_listView_row_activated": gui_tree_row,
   "on_addButton_clicked": gui_btn_add,
   "on_editButton_clicked": gui_btn_edit,
   "on_removeButton_clicked": gui_btn_remove,
   "on_btnUpdate_clicked": gui_btn_update,
   "on_btnDownload_clicked": gui_btn_download,
   "on_btnPlay_clicked": gui_btn_play,
   "on_btnSetup_clicked": gui_btn_setup,
   "on_btnShowAvailable_toggled": gui_btn_show_available,
   "on_btnShowDownloaded_toggled": gui_btn_show_downloaded
 }
glade.signal_autoconnect( connections)

# get treeview to put data into it later on..
tree = glade.get_widget( "listView")
tree.get_selection().set_mode( gtk.SELECTION_MULTIPLE)

# the toggle buttons
btn_showAvailable = glade.get_widget( "btnShowAvailable")
btn_showDownloaded = glade.get_widget( "btnShowDownloaded")

# the download button widgets (and the update button)
img_btn_download = glade.get_widget( "imgDownloadButton")
lbl_btn_download = glade.get_widget( "lblDownloadButton")
btn_update = glade.get_widget( "btnUpdate")

# init some stuff for the gui.. (also load feeds!)
feeds = []
gui_startup()

# enter gtk main loop and give control to gui
gtk.gdk.threads_init()
gtk.main()

saveConfig( configfile)
